import functools
from unittest import TestCase

import psycopg2

from rapt2 import Rapt


class TestTranslator(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = psycopg2.connect(
            database="rapt_test_db", user="raptor", password="raptolicious"
        )
        cls.cur = cls.conn.cursor()
        cls.execute = cls.cur.execute

        cls.setup_database(cls.schema, cls.data)

        cls.translate_bag = staticmethod(
            cls.translate_func(
                functools.partial(
                    Rapt(grammar="Extended Grammar").to_sql, use_bag_semantics=True
                )
            )
        )
        cls.translate_set = staticmethod(
            cls.translate_func(
                functools.partial(Rapt(grammar="Extended Grammar").to_sql)
            )
        )

    @classmethod
    def translate_func(cls, func, schema=None):
        schema = schema or cls.schema
        return functools.partial(func, schema=schema)

    @classmethod
    def tearDownClass(cls):
        cls.drop_tables(cls.schema.keys())

    @classmethod
    def setup_database(cls, schema, data):
        cls.create_tables(schema)
        cls.insert_data(data)

    @classmethod
    def create_tables(cls, schema):
        for relation, attributes in schema.items():
            attributes = ", ".join(
                ["{} text".format(attribute) for attribute in attributes]
            )
            statement = "CREATE TABLE %s (%s)" % (relation, attributes)
            cls.execute(statement)

    @classmethod
    def drop_tables(cls, tables):
        for relation in tables:
            statement = "DROP TABLE %s;" % relation
            cls.execute(statement)

    @classmethod
    def insert_data(cls, data_set):
        for relation, data in data_set.items():
            values = ", ".join(["%s" for _ in data[0]])
            statement = "INSERT INTO {0} VALUES ({1})".format(relation, values)
            for row in data:
                cls.execute(statement, row)

    @classmethod
    def query(cls, statement):
        cls.execute(statement)
        return cls.cur.fetchall()

    @classmethod
    def query_list(cls, statements):
        for statement in statements:
            cls.execute(statement)
        return cls.cur.fetchall()

    @classmethod
    def tearDownClass(cls):
        cls.conn.close()


class TestRelation(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {"Alpha": ["a1", "a2"]}
        cls.data = {"Alpha": [("1", "a"), ("2", "b"), ("2", "b")]}
        super().setUpClass()

    def test_single_relation_set(self):
        instring = "alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a"), ("2", "b")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_relation_set_cased(self):
        instring = "Alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a"), ("2", "b")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_relation_bag(self):
        instring = "alpha;"
        translation = self.translate_bag(instring)
        expected = [("1", "a"), ("2", "b"), ("2", "b")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_multiple_relations_bag(self):
        instring = "alpha; alpha;"
        translation = self.translate_bag(instring)
        expected = [("1", "a"), ("2", "b"), ("2", "b")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestProject(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {"alpha": ["a1", "a2", "a3"]}
        cls.data = {"alpha": [("1", "a", "!"), ("2", "b", "!"), ("3", "b", "!")]}
        super().setUpClass()

    def test_single_project_single_attr_set_no_duplicates(self):
        instring = r"\project_{a1} alpha;"
        translation = self.translate_set(instring)
        expected = [("1",), ("2",), ("3",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_single_attr_set_no_duplicates_cased(self):
        instring = r"\project_{a1} ALPHA;"
        translation = self.translate_set(instring)
        expected = [("1",), ("2",), ("3",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_single_attr_set_with_duplicates(self):
        instring = r"\project_{a2} alpha;"
        translation = self.translate_set(instring)
        expected = [("a",), ("b",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_single_attr_bag_no_duplicates(self):
        instring = r"\project_{a1} alpha;"
        translation = self.translate_bag(instring)
        expected = [("1",), ("2",), ("3",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_single_attr_bag_with_duplicates(self):
        instring = r"\project_{a2} alpha;"
        translation = self.translate_bag(instring)
        expected = [("a",), ("b",), ("b",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_two_attr_set_no_duplicates(self):
        instring = r"\project_{a1, a2} alpha;"
        translation = self.translate_set(instring)
        expected = [
            (
                "1",
                "a",
            ),
            ("2", "b"),
            ("3", "b"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_two_attr_set_with_duplicates(self):
        instring = r"\project_{a2, a3} alpha;"
        translation = self.translate_set(instring)
        expected = [
            (
                "a",
                "!",
            ),
            ("b", "!"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_two_attr_bag_no_duplicates(self):
        instring = r"\project_{a1, a2} alpha;"
        translation = self.translate_bag(instring)
        expected = [
            (
                "1",
                "a",
            ),
            ("2", "b"),
            ("3", "b"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_two_attr_bag_with_duplicates(self):
        instring = r"\project_{a2, a3} alpha;"
        translation = self.translate_bag(instring)
        expected = [
            (
                "a",
                "!",
            ),
            ("b", "!"),
            ("b", "!"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_all_attr(self):
        instring = r"\project_{a1, a2, a3} alpha;"
        translation = self.translate_set(instring)
        expected = self.data["alpha"]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_project_all_attr_out_of_order(self):
        instring = r"\project_{a2, a3, a1} alpha;"
        translation = self.translate_set(instring)
        expected = [("a", "!", "1"), ("b", "!", "2"), ("b", "!", "3")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_double_project(self):
        instring = r"\project_{a2} \project_{a2, a3} alpha;"
        translation = self.translate_set(instring)
        expected = [("a",), ("b",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_triple_project(self):
        instring = (
            r"\project_{a1} \project_{a2, a1}"
            r"\project_{a2, a1, a3} alpha;"
        )
        translation = self.translate_set(instring)
        expected = [("1",), ("2",), ("3",)]
        self.assertCountEqual(expected, self.query_list(translation))


class TestSelect(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {"alpha": ["a1", "a2", "a3"]}
        cls.data = {"alpha": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")]}
        super().setUpClass()

    def test_single_select_with_no_attr_set(self):
        instring = r"\select_{1=1} alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_select_with_no_attr_bag(self):
        instring = r"\select_{1=1} alpha;"
        translation = self.translate_bag(instring)
        expected = self.data["alpha"]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_select_with_attr_all(self):
        instring = r"\select_{a1<>'missing'} alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_select_with_attr_subset_set(self):
        instring = r"\select_{a1<>'1'} alpha;"
        translation = self.translate_set(instring)
        expected = [("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_select_with_attr_subset_bag(self):
        instring = r"\select_{a1<>'1'} alpha;"
        translation = self.translate_bag(instring)
        expected = [("2", "b", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_select_with_attr_none(self):
        instring = r"\select_{a1='1' and a2='b'} alpha;"
        translation = self.translate_set(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_select_with_multiple_cond_attr_subset_set(self):
        instring = r"\select_{a1<>'1' and a2<>'a'} alpha;"
        translation = self.translate_set(instring)
        expected = [("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_single_select_with_attr_case(self):
        instring = r"\select_{A1='1' and A2='b'} Alpha;"
        translation = self.translate_set(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_double_select_with_attr(self):
        instring = r"\select_{a1!='missing'} \select_{A2!='missing'} Alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestRename(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {"alpha": ["a1", "a2", "a3"]}
        cls.data = {"alpha": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")]}
        super().setUpClass()

    def test_rename_relation(self):
        instring = (
            r"\project_{AlphaPrime.a1, AlphaPrime.a2, AlphaPrime.a3} "
            r"\rename_{AlphaPrime} alpha;"
        )
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_rename_relation_bag(self):
        instring = (
            r"\project_{AlphaPrime.a1, AlphaPrime.a2, AlphaPrime.a3} "
            r"\rename_{AlphaPrime} alpha;"
        )
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_rename_attributes(self):
        instring = (
            r"\project_{alpha.ap1, alpha.ap2, alpha.ap3} "
            r"\rename_{(ap1, ap2, ap3)} alpha;"
        )
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_rename_relation_and_attributes(self):
        instring = (
            r"\project_{APrime.ap1, APrime.ap2, APrime.ap3} "
            r"\rename_{APrime(ap1, ap2, ap3)} alpha;"
        )
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestAssign(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {"alpha": ["a1", "a2", "a3"]}
        cls.data = {"alpha": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")]}
        super().setUpClass()

    def tearDown(self):
        self.drop_tables(["alpha_prime"])

    def test_assign_relation_set(self):
        instring = "alpha_prime := alpha; alpha_prime;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_assign_relation_bag(self):
        instring = "alpha_prime := alpha; alpha_prime;"
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_assign_relation_with_attr_rename(self):
        instring = (
            r"alpha_prime(ap1, ap2, ap3) := alpha;"
            r"\project_{alpha_prime.ap1, alpha_prime.ap2, alpha_prime.ap3} alpha_prime;"
        )
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestUnion(TestTranslator):
    @classmethod
    def setUpClass(cls):
        attrs = ["a1", "a2", "a3"]
        cls.schema = {
            "alpha": attrs,
            "alpha_copy": attrs,
            "alpha_subset": attrs,
            "alpha_superset": attrs,
            "alpha_extra": attrs,
        }
        cls.data = {
            "alpha": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")],
            "alpha_copy": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")],
            "alpha_subset": [
                ("1", "a", "!"),
            ],
            "alpha_superset": [
                ("1", "a", "!"),
                ("2", "b", "!"),
                ("2", "b", "!"),
                ("3", "c", "?"),
            ],
            "alpha_extra": [
                ("3", "c", "?"),
            ],
        }
        super().setUpClass()

    def test_union_self_set(self):
        instring = "alpha \\union alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_self_bag(self):
        instring = "alpha \\union alpha;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "!"),
            ("2", "b", "!"),
            ("2", "b", "!"),
            ("1", "a", "!"),
            ("2", "b", "!"),
            ("2", "b", "!"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_other_set(self):
        instring = "alpha \\union alpha_copy;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_other_bag(self):
        instring = "alpha \\union alpha_copy;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "!"),
            ("2", "b", "!"),
            ("2", "b", "!"),
            ("1", "a", "!"),
            ("2", "b", "!"),
            ("2", "b", "!"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_superset_set(self):
        instring = "alpha \\union alpha_superset;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_superset_bag(self):
        instring = "alpha \\union alpha_superset;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "!"),
            ("2", "b", "!"),
            ("2", "b", "!"),
            ("1", "a", "!"),
            ("2", "b", "!"),
            ("2", "b", "!"),
            ("3", "c", "?"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_extra_set(self):
        instring = "alpha \\union alpha_extra;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_extra_bag(self):
        instring = "alpha \\union alpha_extra;"
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!"), ("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_subset_set(self):
        instring = "alpha \\union alpha_subset;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_union_subset_bag(self):
        instring = "alpha \\union alpha_extra;"
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!"), ("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestDifference(TestTranslator):
    @classmethod
    def setUpClass(cls):
        attrs = ["a1", "a2", "a3"]
        cls.schema = {
            "alpha": attrs,
            "alpha_copy": attrs,
            "alpha_subset": attrs,
            "alpha_superset": attrs,
            "alpha_extra": attrs,
        }
        cls.data = {
            "alpha": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")],
            "alpha_copy": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")],
            "alpha_subset": [
                ("1", "a", "!"),
            ],
            "alpha_superset": [
                ("1", "a", "!"),
                ("2", "b", "!"),
                ("2", "b", "!"),
                ("3", "c", "?"),
                ("3", "c", "?"),
            ],
            "alpha_extra": [
                ("3", "c", "?"),
                ("3", "c", "?"),
            ],
        }
        super().setUpClass()

    def test_difference_self_set(self):
        instring = "alpha \\difference alpha;"
        translation = self.translate_set(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_self_bag(self):
        instring = "alpha \\difference alpha;"
        translation = self.translate_bag(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_other_set(self):
        instring = "alpha \\difference alpha_copy;"
        translation = self.translate_set(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_other_bag(self):
        instring = "alpha \\difference alpha_copy;"
        translation = self.translate_set(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_subset_set(self):
        instring = "alpha_subset \\difference alpha;"
        translation = self.translate_set(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_subset_bag(self):
        instring = "alpha_subset \\difference alpha;"
        translation = self.translate_bag(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_superset_set(self):
        instring = "alpha_superset \\difference alpha;"
        translation = self.translate_set(instring)
        expected = [("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_superset_bag(self):
        instring = "alpha_superset \\difference alpha;"
        translation = self.translate_bag(instring)
        expected = [("3", "c", "?"), ("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_extra_set(self):
        instring = "alpha_extra \\difference alpha;"
        translation = self.translate_set(instring)
        expected = [("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_difference_extra_bag(self):
        instring = "alpha_extra \\difference alpha;"
        translation = self.translate_bag(instring)
        expected = [("3", "c", "?"), ("3", "c", "?")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestIntersection(TestTranslator):
    @classmethod
    def setUpClass(cls):
        attrs = ["a1", "a2", "a3"]
        cls.schema = {
            "alpha": attrs,
            "alpha_copy": attrs,
            "alpha_subset": attrs,
            "alpha_superset": attrs,
            "alpha_extra": attrs,
        }
        cls.data = {
            "alpha": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")],
            "alpha_copy": [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")],
            "alpha_subset": [
                ("1", "a", "!"),
            ],
            "alpha_superset": [
                ("1", "a", "!"),
                ("2", "b", "!"),
                ("2", "b", "!"),
                ("3", "c", "?"),
                ("3", "c", "?"),
            ],
            "alpha_extra": [
                ("3", "c", "?"),
                ("3", "c", "?"),
            ],
        }
        super().setUpClass()

    def test_intersect_self_set(self):
        instring = "alpha \\intersect alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_self_bag(self):
        instring = "alpha \\intersect alpha;"
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_other_set(self):
        instring = "alpha \\intersect alpha_copy;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_other_bag(self):
        instring = "alpha \\intersect alpha_copy;"
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!"), ("2", "b", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_superset_set(self):
        instring = "alpha \\intersect alpha_superset;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_superset_bag(self):
        instring = "alpha \\intersect alpha_superset;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "!"),
            ("2", "b", "!"),
            ("2", "b", "!"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_extra_set(self):
        instring = "alpha_extra \\intersect alpha;"
        translation = self.translate_set(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_extra_bag(self):
        instring = "alpha_extra \\intersect alpha;"
        translation = self.translate_bag(instring)
        expected = []
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_subset_set(self):
        instring = "alpha \\intersect alpha_subset;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_intersect_subset_bag(self):
        instring = "alpha \\intersect alpha_subset;"
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestJoin(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2"],
            "beta": ["b1"],
            "gamma": ["g1"],
        }
        cls.data = {
            "alpha": [("1", "a"), ("2", "b"), ("2", "b")],
            "beta": [
                ("3",),
                ("4",),
            ],
            "gamma": [("?",), ("!",)],
        }
        super().setUpClass()

    def test_join_other_set(self):
        instring = "alpha \\join beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "3"),
            ("2", "b", "3"),
            ("1", "a", "4"),
            ("2", "b", "4"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_join_other_bag(self):
        instring = "alpha \\join beta;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "3"),
            ("2", "b", "3"),
            ("2", "b", "3"),
            ("1", "a", "4"),
            ("2", "b", "4"),
            ("2", "b", "4"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_join_double(self):
        instring = "alpha \\join beta \\join gamma;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "3", "?"),
            ("2", "b", "3", "?"),
            ("1", "a", "4", "?"),
            ("2", "b", "4", "?"),
            ("1", "a", "3", "!"),
            ("2", "b", "3", "!"),
            ("1", "a", "4", "!"),
            ("2", "b", "4", "!"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_join_self_with_rename_set(self):
        instring = "\\rename_{a1} alpha \\join \\rename_{a2} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "a"),
            ("2", "b", "1", "a"),
            ("1", "a", "2", "b"),
            ("2", "b", "2", "b"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))


class TestNaturalJoin(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2"],
            "alpha_copy": ["a1", "a2"],
            "alpha_subset": ["a1", "a2"],
            "alpha_prime": ["a1", "ap2"],
            "alpha_prime_extra": ["ap2", "ap3"],
        }
        cls.data = {
            "alpha": [("1", "a"), ("2", "b"), ("2", "b")],
            "alpha_copy": [("1", "a"), ("2", "b"), ("2", "b")],
            "alpha_subset": [("2", "b"), ("2", "b")],
            "alpha_prime": [("1", "!"), ("2", "?")],
            "alpha_prime_extra": [("?", "this")],
        }
        super().setUpClass()

    def test_join_copy_set(self):
        instring = "alpha \\natural_join alpha_copy;"
        translation = self.translate_set(instring)
        expected = [("1", "a"), ("2", "b")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_join_copy_bag(self):
        instring = "alpha \\natural_join alpha_copy;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a"),
            ("2", "b"),
            ("2", "b"),
            ("1", "a"),
            ("2", "b"),
            ("2", "b"),
        ]
        self.assertCountEqual(set(expected), set(self.query_list(translation)))

    def test_join_prime_set(self):
        instring = "alpha \\natural_join alpha_prime;"
        translation = self.translate_set(instring)
        expected = [("1", "a", "!"), ("2", "b", "?")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_join_prime_bag(self):
        instring = "alpha \\natural_join alpha_prime;"
        translation = self.translate_bag(instring)
        expected = [("1", "a", "!"), ("2", "b", "?"), ("2", "b", "?")]
        self.assertCountEqual(set(expected), set(self.query_list(translation)))

    def test_join_prime_extra_set(self):
        instring = "alpha \\natural_join alpha_prime \\natural_join alpha_prime_extra;"
        translation = self.translate_set(instring)
        expected = [("2", "b", "?", "this")]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_join_prime_extra_bag(self):
        instring = "alpha \\natural_join alpha_prime \\natural_join alpha_prime_extra;"
        translation = self.translate_bag(instring)
        expected = [
            ("2", "b", "?", "this"),
            ("2", "b", "?", "this"),
        ]
        self.assertCountEqual(set(expected), set(self.query_list(translation)))

    def test_join_self_with_rename_set(self):
        instring = "\\rename_{a1} alpha \\natural_join \\rename_{a2} alpha;"
        translation = self.translate_set(instring)
        expected = [("1", "a"), ("2", "b")]
        self.assertCountEqual(expected, self.query_list(translation))


class TestThetaJoin(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2"],
            "alpha_copy": ["a1", "a2"],
            "alpha_prime": ["ap1", "ap2"],
            "alpha_prime_subset": ["ap1", "ap2"],
        }
        cls.data = {
            "alpha": [("1", "a"), ("2", "b"), ("2", "b")],
            "alpha_copy": [("1", "a"), ("2", "b"), ("2", "b")],
            "alpha_prime": [("1", "a"), ("2", "b"), ("2", "b")],
            "alpha_prime_subset": [("1", "!")],
        }
        super().setUpClass()

    def test_join_copy_set(self):
        instring = "alpha \\join_{alpha.a1=alpha_copy.a1} alpha_copy;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "a"),
            ("2", "b", "2", "b"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_join_copy_bag(self):
        instring = "alpha \\join_{alpha.a1=alpha_copy.a1} alpha_copy;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "1", "a"),
            ("2", "b", "2", "b"),
            ("2", "b", "2", "b"),
        ]
        self.assertCountEqual(set(expected), set(self.query_list(translation)))

    def test_join_prime_set(self):
        instring = "alpha \\join_{alpha.a1=alpha_prime.ap1} alpha_prime;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "a"),
            ("2", "b", "2", "b"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))


class TestUnary(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {"Alpha": ["a1", "a2"]}
        cls.data = {"Alpha": [("1", "a"), ("2", "b"), ("2", "b")]}
        super().setUpClass()

    def test_select_project(self):
        instring = r"\select_{a2='b'} \project_{a2 } alpha;"
        translation = self.translate_set(instring)
        expected = [("b",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_project_select(self):
        instring = r" \project_{a2} \select_{a2='b'} alpha;"
        translation = self.translate_set(instring)
        expected = [("b",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_select_rename(self):
        instring = r"\select_{ap2='b'} \rename_{A(ap1, ap2)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            (
                "2",
                "b",
            )
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_project_rename(self):
        instring = r"\project_{ap2} \rename_{A(ap1, ap2)} alpha;"
        translation = self.translate_set(instring)
        expected = [("a",), ("b",)]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_assign_project_rename(self):
        instring = r"A := \project_{ap2} \rename_{A(ap1, ap2)} alpha; A;"
        translation = self.translate_set(instring)
        expected = [("a",), ("b",)]
        self.assertCountEqual(expected, self.query_list(translation))


class TestFullOuterJoin(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2"],
            "beta": ["b1", "b2"],
            "gamma": ["g1", "g2"],
        }
        cls.data = {
            "alpha": [("1", "a"), ("2", "b"), ("3", "c")],
            "beta": [("1", "x"), ("2", "y"), ("4", "z")],
            "gamma": [("5", "p"), ("6", "q")],
        }
        super().setUpClass()

    def test_full_outer_join_basic_set(self):
        instring = "alpha \\full_outer_join_{a1 = b1} beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", "2", "y"),
            ("3", "c", None, None),
            (None, None, "4", "z"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_full_outer_join_basic_bag(self):
        instring = "alpha \\full_outer_join_{a1 = b1} beta;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", "2", "y"),
            ("3", "c", None, None),
            (None, None, "4", "z"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_full_outer_join_complex_condition_set(self):
        instring = "alpha \\full_outer_join_{alpha.a1 = beta.b1 and a2 = 'a'} beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", None, None),
            ("3", "c", None, None),
            (None, None, "2", "y"),
            (None, None, "4", "z"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_full_outer_join_with_project_set(self):
        instring = (
            "\\project_{alpha.a1, beta.b1} (alpha \\full_outer_join_{a1 = b1} beta);"
        )
        translation = self.translate_set(instring)
        expected = [
            ("1", "1"),
            ("2", "2"),
            ("3", None),
            (None, "4"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_full_outer_join_with_select_set(self):
        instring = "\\select_{alpha.a1 > '1'} (alpha \\full_outer_join_{a1 = b1} beta);"
        translation = self.translate_set(instring)
        expected = [
            ("2", "b", "2", "y"),
            ("3", "c", None, None),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_full_outer_join_chained_set(self):
        instring = "(alpha \\full_outer_join_{a1 = b1} beta) \\join gamma;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x", "5", "p"),
            ("1", "a", "1", "x", "6", "q"),
            ("2", "b", "2", "y", "5", "p"),
            ("2", "b", "2", "y", "6", "q"),
            ("3", "c", None, None, "5", "p"),
            ("3", "c", None, None, "6", "q"),
            (None, None, "4", "z", "5", "p"),
            (None, None, "4", "z", "6", "q"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))


class TestLeftOuterJoin(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2"],
            "beta": ["b1", "b2"],
            "gamma": ["g1", "g2"],
        }
        cls.data = {
            "alpha": [("1", "a"), ("2", "b"), ("3", "c")],
            "beta": [("1", "x"), ("2", "y"), ("4", "z")],
            "gamma": [("5", "p"), ("6", "q")],
        }
        super().setUpClass()

    def test_left_outer_join_basic_set(self):
        instring = "alpha \\left_outer_join_{a1 = b1} beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", "2", "y"),
            ("3", "c", None, None),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_left_outer_join_basic_bag(self):
        instring = "alpha \\left_outer_join_{a1 = b1} beta;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", "2", "y"),
            ("3", "c", None, None),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_left_outer_join_complex_condition_set(self):
        instring = "alpha \\left_outer_join_{alpha.a1 = beta.b1 and a2 = 'a'} beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", None, None),
            ("3", "c", None, None),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_left_outer_join_with_project_set(self):
        instring = (
            "\\project_{alpha.a1, beta.b1} (alpha \\left_outer_join_{a1 = b1} beta);"
        )
        translation = self.translate_set(instring)
        expected = [
            ("1", "1"),
            ("2", "2"),
            ("3", None),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_left_outer_join_with_select_set(self):
        instring = "\\select_{alpha.a1 > '1'} (alpha \\left_outer_join_{a1 = b1} beta);"
        translation = self.translate_set(instring)
        expected = [
            ("2", "b", "2", "y"),
            ("3", "c", None, None),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_left_outer_join_chained_set(self):
        instring = "(alpha \\left_outer_join_{a1 = b1} beta) \\join gamma;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x", "5", "p"),
            ("1", "a", "1", "x", "6", "q"),
            ("2", "b", "2", "y", "5", "p"),
            ("2", "b", "2", "y", "6", "q"),
            ("3", "c", None, None, "5", "p"),
            ("3", "c", None, None, "6", "q"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))


class TestRightOuterJoin(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2"],
            "beta": ["b1", "b2"],
            "gamma": ["g1", "g2"],
        }
        cls.data = {
            "alpha": [("1", "a"), ("2", "b"), ("3", "c")],
            "beta": [("1", "x"), ("2", "y"), ("4", "z")],
            "gamma": [("5", "p"), ("6", "q")],
        }
        super().setUpClass()

    def test_right_outer_join_basic_set(self):
        instring = "alpha \\right_outer_join_{a1 = b1} beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", "2", "y"),
            (None, None, "4", "z"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_right_outer_join_basic_bag(self):
        instring = "alpha \\right_outer_join_{a1 = b1} beta;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "1", "x"),
            ("2", "b", "2", "y"),
            (None, None, "4", "z"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_right_outer_join_complex_condition_set(self):
        instring = "alpha \\right_outer_join_{alpha.a1 = beta.b1 and a2 = 'a'} beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x"),
            (None, None, "2", "y"),
            (None, None, "4", "z"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_right_outer_join_with_project_set(self):
        instring = (
            "\\project_{alpha.a1, beta.b1} (alpha \\right_outer_join_{a1 = b1} beta);"
        )
        translation = self.translate_set(instring)
        expected = [
            ("1", "1"),
            ("2", "2"),
            (None, "4"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_right_outer_join_with_select_set(self):
        instring = "\\select_{beta.b1 > '1'} (alpha \\right_outer_join_{a1 = b1} beta);"
        translation = self.translate_set(instring)
        expected = [
            ("2", "b", "2", "y"),
            (None, None, "4", "z"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_right_outer_join_chained_set(self):
        instring = "(alpha \\right_outer_join_{a1 = b1} beta) \\join gamma;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "1", "x", "5", "p"),
            ("1", "a", "1", "x", "6", "q"),
            ("2", "b", "2", "y", "5", "p"),
            ("2", "b", "2", "y", "6", "q"),
            (None, None, "4", "z", "5", "p"),
            (None, None, "4", "z", "6", "q"),
        ]
        self.assertCountEqual(expected, self.query_list(translation))


class TestDefined(TestTranslator):
    @classmethod
    def setUpClass(cls):
        cls.schema = {
            "alpha": ["a1", "a2", "a3"],
            "beta": ["b1", "b2"],
        }
        cls.data = {
            "alpha": [
                ("1", "a", "x"),  # All non-null
                ("2", None, "y"),  # a2 is null
                (None, "c", "z"),  # a1 is null
                ("4", "d", None),  # a3 is null
                (None, None, None),  # All null
            ],
            "beta": [
                ("1", "p"),  # All non-null
                (None, "q"),  # b1 is null
                ("3", None),  # b2 is null
            ],
        }
        super().setUpClass()

    def test_defined_single_attribute_set(self):
        instring = "\\select_{defined(a1)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # a1 = "1" (defined)
            ("2", None, "y"),  # a1 = "2" (defined)
            ("4", "d", None),  # a1 = "4" (defined)
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_single_attribute_bag(self):
        instring = "\\select_{defined(a1)} alpha;"
        translation = self.translate_bag(instring)
        expected = [
            ("1", "a", "x"),  # a1 = "1" (defined)
            ("2", None, "y"),  # a1 = "2" (defined)
            ("4", "d", None),  # a1 = "4" (defined)
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_not_defined_single_attribute_set(self):
        instring = "\\select_{not defined(a1)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            (None, "c", "z"),  # a1 is null
            (None, None, None),  # a1 is null
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_multiple_attributes_set(self):
        instring = "\\select_{defined(a1) and defined(a2)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # Both a1 and a2 defined
            ("4", "d", None),  # Both a1 and a2 defined (a3 is null but that's OK)
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_or_defined_set(self):
        instring = "\\select_{defined(a1) or defined(a2)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # Both defined
            ("2", None, "y"),  # a1 defined, a2 null
            (None, "c", "z"),  # a1 null, a2 defined
            ("4", "d", None),  # a1 defined, a2 defined
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_with_equality_set(self):
        instring = "\\select_{defined(a1) and a1 = '1'} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # a1 defined and equals "1"
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_not_defined_with_equality_set(self):
        instring = "\\select_{not defined(a1) or a2 = 'c'} alpha;"
        translation = self.translate_set(instring)
        expected = [
            (None, "c", "z"),  # a1 not defined, a2 = "c"
            (None, None, None),  # a1 not defined
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_complex_defined_condition_set(self):
        instring = "\\select_{defined(a1) and (a2 = 'a' or not defined(a3))} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # a1 defined, a2 = "a"
            ("4", "d", None),  # a1 defined, a3 not defined
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_with_relation_attribute_set(self):
        instring = "\\select_{defined(alpha.a1)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # alpha.a1 defined
            ("2", None, "y"),  # alpha.a1 defined
            ("4", "d", None),  # alpha.a1 defined
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_with_project_set(self):
        instring = "\\project_{a1, a2} \\select_{defined(a1)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a"),  # a1 defined
            ("2", None),  # a1 defined
            ("4", "d"),  # a1 defined
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_in_join_condition_set(self):
        instring = "alpha \\join_{defined(alpha.a1) and defined(beta.b1)} beta;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x", "1", "p"),  # Both a1 and b1 defined
            ("2", None, "y", "1", "p"),  # Both a1 and b1 defined
            ("4", "d", None, "1", "p"),  # Both a1 and b1 defined
            ("1", "a", "x", "3", None),  # Both a1 and b1 defined (b1="3")
            ("2", None, "y", "3", None),  # Both a1 and b1 defined (b1="3")
            ("4", "d", None, "3", None),  # Both a1 and b1 defined (b1="3")
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_with_outer_join_set(self):
        instring = (
            "alpha \\left_outer_join_{defined(alpha.a1) and defined(beta.b1)} beta;"
        )
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x", "1", "p"),  # Both defined - matches
            ("2", None, "y", "1", "p"),  # Both defined - matches
            ("4", "d", None, "1", "p"),  # Both defined - matches
            ("1", "a", "x", "3", None),  # Both defined - matches (b1="3")
            ("2", None, "y", "3", None),  # Both defined - matches (b1="3")
            ("4", "d", None, "3", None),  # Both defined - matches (b1="3")
            (None, "c", "z", None, None),  # a1 not defined - no match
            (None, None, None, None, None),  # a1 not defined - no match
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_with_union_set(self):
        instring = "\\select_{defined(a1)} alpha \\union \\select_{defined(a1)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # From alpha where a1 defined
            ("2", None, "y"),  # From alpha where a1 defined
            ("4", "d", None),  # From alpha where a1 defined
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_multiple_defined_conditions_set(self):
        instring = "\\select_{defined(a1)} \\select_{defined(a2)} alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # Both a1 and a2 defined
            ("4", "d", None),  # Both a1 and a2 defined (a3 is null)
        ]
        self.assertCountEqual(expected, self.query_list(translation))

    def test_defined_with_assign_set(self):
        instring = "defined_alpha := \\select_{defined(a1)} alpha; defined_alpha;"
        translation = self.translate_set(instring)
        expected = [
            ("1", "a", "x"),  # a1 defined
            ("2", None, "y"),  # a1 defined
            ("4", "d", None),  # a1 defined
        ]
        self.assertCountEqual(expected, self.query_list(translation))



import unittest


class GrammarTestCase(unittest.TestCase):
    @staticmethod
    def parse_function(parser):
        """
        Return a function that parses a value with the specified parser and
        returns the results as a list.
        """

        def function(value):
            return parser.parseString(value, parseAll=True).asList()

        return function



from typing import TypeVar

from pyparsing import CaselessKeyword, Group, Literal, infixNotation, opAssoc

from .core_grammar import CoreGrammar
from .syntax import ExtendedSyntax

TExtendedSyntax = TypeVar("TExtendedSyntax", bound=ExtendedSyntax)


class ExtendedGrammar(CoreGrammar[TExtendedSyntax]):
    """
    A parser that recognizes an extended relational algebra grammar.

    The rules are annotated with their BNF equivalents. For a complete
    specification refer to the associated grammar file.
    """

    syntax: TExtendedSyntax

    def __init__(self, syntax: TExtendedSyntax = ExtendedSyntax()):
        """
        Initializes an ExtendedGrammar. Uses the default syntax if none
        is provided.

        :param syntax: a syntax for this grammar.
        """
        super().__init__(syntax)

    @property
    def natural_join(self):
        """
        natural_join_expr ::= expression natural_join expression
        """
        return CaselessKeyword(self.syntax.natural_join_op)

    @property
    def theta_join(self):
        """
        select_expr ::= select param_start conditions param_stop expression
        """
        long = self.parametrize(self.syntax.theta_join_op, self.conditions)
        short = self.parametrize(self.syntax.join_op, self.conditions).setParseAction(
            self.theta_parse_action
        )
        return long ^ short

    def theta_parse_action(self, s, loc, t):
        t[0] = self.syntax.theta_join_op
        return t

    @property
    def full_outer_join(self):
        """
        full_outer_join_expr ::= expression \\full_outer_join param_start conditions param_stop expression
        """
        return self.parametrize(self.syntax.full_outer_join_op, self.conditions)

    @property
    def left_outer_join(self):
        """
        left_outer_join_expr ::= expression \\left_outer_join param_start conditions param_stop expression
        """
        return self.parametrize(self.syntax.left_outer_join_op, self.conditions)

    @property
    def right_outer_join(self):
        """
        right_outer_join_expr ::= expression \\right_outer_join param_start conditions param_stop expression
        """
        return self.parametrize(self.syntax.right_outer_join_op, self.conditions)

    @property
    def binary_op_p1(self):
        return (
            super().binary_op_p1
            ^ self.natural_join
            ^ self.theta_join
            ^ self.full_outer_join
            ^ self.left_outer_join
            ^ self.right_outer_join
        )

    @property
    def intersect(self):
        """
        intersect_op ::= intersect_name
        """
        return CaselessKeyword(self.syntax.intersect_op)

    @property
    def defined_op(self):
        return CaselessKeyword(self.syntax.defined_op)

    @property
    def defined_condition(self):
        """
        defined_condition ::= defined_op "(" operand ")"
        """
        return Group(
            self.defined_op
            + Literal(self.syntax.paren_left).suppress()
            + self.operand
            + Literal(self.syntax.paren_right).suppress()
        )

    @property
    def condition(self):
        """
        condition ::= condition | defined_condition
        """
        return super().condition | self.defined_condition

    @property
    def expression(self):
        return infixNotation(
            self.relation,
            [
                (self.unary_op, 1, opAssoc.RIGHT),
                (self.binary_op_p1, 2, opAssoc.LEFT),
                (self.intersect, 2, opAssoc.LEFT),
                (self.binary_op_p2, 2, opAssoc.LEFT),
            ],
            lpar=self.syntax.paren_left,
            rpar=self.syntax.paren_right,
        )

    def is_unary(self, operator):
        return operator in {
            self.syntax.select_op,
            self.syntax.project_op,
            self.syntax.rename_op,
        }

    def is_binary(self, operator):
        return operator in {
            self.syntax.intersect_op,
            self.syntax.natural_join_op,
            self.syntax.theta_join_op,
            self.syntax.full_outer_join_op,
            self.syntax.left_outer_join_op,
            self.syntax.right_outer_join_op,
        } or super().is_binary(operator)



from unittest import TestCase

from rapt2.treebrd.grammars.syntax import Syntax


class TestSyntax(TestCase):
    def test___init__when_operator_is_changed(self):
        new_op = "test_op"
        self.assertEqual(new_op, Syntax(and_op=new_op).and_op)

    def test___init__cannot_set_unknown_attributes(self):
        self.assertFalse(hasattr(Syntax(foo="bar"), "foo"))



from ...treebrd.node import Operator
from ...treebrd.condition_node import BinaryConditionalOperator, UnaryConditionalOperator
from .constants import (
    SELECT_OP, PROJECT_OP, RENAME_OP, ASSIGN_OP, JOIN_OP, NATURAL_JOIN_OP, THETA_JOIN_OP,
    FULL_OUTER_JOIN_OP, LEFT_OUTER_JOIN_OP, RIGHT_OUTER_JOIN_OP, UNION_OP, DIFFERENCE_OP,
    INTERSECT_OP, PRIMARY_KEY_OP, MULTIVALUED_DEPENDENCY_OP, FUNCTIONAL_DEPENDENCY_OP,
    INCLUSION_EQUIVALENCE_OP, INCLUSION_SUBSUMPTION_OP, AND_OP, OR_OP, EQUAL_OP,
    NOT_EQUAL_OP, LESS_THAN_OP, LESS_THAN_EQUAL_OP, GREATER_THAN_OP, GREATER_THAN_EQUAL_OP,
    NOT_OP, DEFINED_OP
)

latex_operator = {
    Operator.select: SELECT_OP,
    Operator.project: PROJECT_OP,
    Operator.rename: RENAME_OP,
    Operator.assign: ASSIGN_OP,
    Operator.cross_join: JOIN_OP,
    Operator.natural_join: NATURAL_JOIN_OP,
    Operator.theta_join: THETA_JOIN_OP,
    Operator.full_outer_join: FULL_OUTER_JOIN_OP,
    Operator.left_outer_join: LEFT_OUTER_JOIN_OP,
    Operator.right_outer_join: RIGHT_OUTER_JOIN_OP,
    Operator.union: UNION_OP,
    Operator.difference: DIFFERENCE_OP,
    Operator.intersect: INTERSECT_OP,
    Operator.primary_key: PRIMARY_KEY_OP,
    Operator.multivalued_dependency: MULTIVALUED_DEPENDENCY_OP,
    Operator.functional_dependency: FUNCTIONAL_DEPENDENCY_OP,
    Operator.inclusion_equivalence: INCLUSION_EQUIVALENCE_OP,
    Operator.inclusion_subsumption: INCLUSION_SUBSUMPTION_OP,
}

# Conditional operator mappings
conditional_latex_operator = {
    # Binary operators
    BinaryConditionalOperator.AND: AND_OP,
    BinaryConditionalOperator.OR: OR_OP,
    BinaryConditionalOperator.EQUAL: EQUAL_OP,
    BinaryConditionalOperator.NOT_EQUAL: NOT_EQUAL_OP,
    BinaryConditionalOperator.LESS_THAN: LESS_THAN_OP,
    BinaryConditionalOperator.LESS_THAN_EQUAL: LESS_THAN_EQUAL_OP,
    BinaryConditionalOperator.GREATER_THAN: GREATER_THAN_OP,
    BinaryConditionalOperator.GREATER_THAN_EQUAL: GREATER_THAN_EQUAL_OP,
    # Unary operators
    UnaryConditionalOperator.NOT: NOT_OP,
    UnaryConditionalOperator.DEFINED: DEFINED_OP,
}



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestUnion(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_binary_precedence_low_high(self):
        expression = "astronauts \\union sputnik1_crew \\join sputnik2_crew;"
        expected = [
            [
                ["astronauts"],
                "\\union",
                [["sputnik1_crew"], "\\join", ["sputnik2_crew"]],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_binary_precedence_high_low(self):
        expression = "astronauts \\join sputnik1_crew \\union sputnik2_crew;"
        expected = [
            [
                [["astronauts"], "\\join", ["sputnik1_crew"]],
                "\\union",
                ["sputnik2_crew"],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_binary_precedence_high_low_bracketed(self):
        expression = "astronauts \\join (sputnik1_crew \\union sputnik2_crew);"
        expected = [
            [
                ["astronauts"],
                "\\join",
                [["sputnik1_crew"], "\\union", ["sputnik2_crew"]],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_binary_same_precedence(self):
        expression = "astronauts \\union s1_crew \\difference s2_crew;"
        expected = [
            [["astronauts"], "\\union", ["s1_crew"], "\\difference", ["s2_crew"]]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_basic(self):
        expression = "\\union;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_relation(self):
        expression = "\\union_{species};"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_attributes(self):
        expression = "\\union astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_empty_attribute_list(self):
        expression = "astronauts \\union;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )



from unittest import TestCase

from rapt2.treebrd.attributes import Attribute, AttributeList

__author__ = "Noel"


class TestAttribute(TestCase):
    def test_name_after_init(self):
        attribute = Attribute("Name", "Prefix")
        self.assertEqual("Name", attribute.name)

    def test_prefix_after_init(self):
        attribute = Attribute("Name", "Prefix")
        self.assertEqual("Prefix", attribute.prefix)

    def test_prefixed_after_init(self):
        attribute = Attribute("Name", "Prefix")
        self.assertEqual("Prefix.Name", attribute.prefixed)

    def test_equality_when_name_and_attribute_are_equal(self):
        attribute_a = Attribute("Name", "Prefix")
        attribute_b = Attribute("Name", "Prefix")
        self.assertEqual(attribute_a, attribute_b)

    def test_equality_when_name_and_attribute_are_different(self):
        attribute_a = Attribute("Name", "Prefix")
        attribute_b = Attribute("Other", "Other")
        self.assertNotEqual(attribute_a, attribute_b)

    def test_equality_when_name_is_different(self):
        attribute_a = Attribute("Name", "Prefix")
        attribute_b = Attribute("Other", "Prefix")
        self.assertNotEqual(attribute_a, attribute_b)

    def test_equality_when_attribute_is_different(self):
        attribute_a = Attribute("Name", "Prefix")
        attribute_b = Attribute("Name", "Other")
        self.assertNotEqual(attribute_a, attribute_b)

    def test_hash_when_name_and_attribute_are_equal(self):
        attribute_a = Attribute("Name", "Prefix")
        attribute_b = Attribute("Name", "Prefix")
        self.assertEqual(hash(attribute_a), hash(attribute_b))


class TestAttributeList(TestCase):
    def test_trim_when_restriction_is_empty(self):
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim([])
        self.assertEqual([], a_list.to_list())

    def test_trim_when_restriction_contains_all_with_no_prefix(self):
        expected = ["prefix.A", "prefix.B", "prefix.C"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["A", "B", "C"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_restriction_contains_all_with_prefix(self):
        expected = ["prefix.A", "prefix.B", "prefix.C"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["prefix.A", "prefix.B", "prefix.C"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_restriction_excludes_first(self):
        expected = ["prefix.B", "prefix.C"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["B", "C"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_restriction_excludes_last(self):
        expected = ["prefix.A", "prefix.B"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["A", "B"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_restriction_excludes_middle(self):
        expected = ["prefix.A", "prefix.C"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["A", "C"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_restriction_leaves_first(self):
        expected = ["prefix.A"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["A"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_restriction_leaves_last(self):
        expected = ["prefix.C"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["C"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_restriction_leaves_middle(self):
        expected = ["prefix.B"]
        a_list = AttributeList(["A", "B", "C"], "prefix")
        a_list.trim(["B"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_reorder(self):
        expected = ["prefix.B", "prefix.C", "prefix.A"]
        a_list = AttributeList(["B", "C", "A"], "prefix")
        a_list.trim(["B", "C", "A"])
        self.assertEqual(expected, a_list.to_list())

    def test_trim_when_reorder_and_restrict(self):
        expected = ["prefix.C", "prefix.A"]
        a_list = AttributeList(["B", "C", "A"], "prefix")
        a_list.trim(["C", "A"])
        self.assertEqual(expected, a_list.to_list())

    def test_rename_prefix(self):
        expected = ["prefix.A", "prefix.B", "prefix.C"]
        a_list = AttributeList(["A", "B", "C"], "old")
        a_list.rename([], "prefix")
        self.assertEqual(expected, a_list.to_list())

    def test_rename_attribute_names(self):
        expected = ["prefix.A", "prefix.B", "prefix.C"]
        a_list = AttributeList(["a", "b", "c"], "prefix")
        a_list.rename(["A", "B", "C"], None)
        self.assertEqual(expected, a_list.to_list())

    def test_rename_attribute_name_and_prefix(self):
        expected = ["prefix.A", "prefix.B", "prefix.C"]
        a_list = AttributeList(["a", "b", "c"], "old")
        a_list.rename(["A", "B", "C"], "prefix")
        self.assertEqual(expected, a_list.to_list())



from .transformers.qtree import qtree_translator
from .transformers.sql import sql_translator
from .treebrd.grammars import GRAMMARS, CoreGrammar
from .treebrd.grammars.syntax import Syntax
from .treebrd.treebrd import TreeBRD


class Rapt:
    @staticmethod
    def configure_grammar(**config):
        syntax = Syntax(**config.get("syntax", {}))
        grammar_class_name = config.get("grammar", "Extended Grammar")
        grammar_class = GRAMMARS.get(grammar_class_name, CoreGrammar)
        return grammar_class(syntax)

    def __init__(self, **config):
        grammar = self.configure_grammar(**config)
        self.builder = TreeBRD(grammar)

    def to_syntax_tree(self, instring, schema={}):
        """
        Return a list of syntax trees that represent the instring.

        :param instring: a relational algebra string
        :param schema: a schema for the string
        :return: a list of syntax trees
        """
        return self.builder.build(instring, schema)

    def to_sql(self, instring, schema={}, use_bag_semantics=False):
        """
        Translate a relational algebra string into a SQL string.

        :param instring: a relational algebra string to translate
        :param schema: a mapping of relation names to their attributes
        :param use_bag_semantics: flag for using relational algebra bag semantics
        :return: a SQL translation string
        """
        root_list = self.to_syntax_tree(instring, schema)
        return sql_translator.translate(
            root_list, use_bag_semantics, self.builder.grammar.syntax
        )  # type: ignore

    def to_sql_sequence(self, instring, schema={}, use_bag_semantics=False):
        """
        Translate a relational algebra string into a list of SQL strings generated
        by a post-order traversal of the parse tree for the input string.

        :param instring: a relational algebra string to translate
        :param schema: a mapping of relation names to their attributes
        :param use_bag_semantics: flag for using relational algebra bag semantics
        :return: a list of SQL translation strings
        """
        root_list = self.to_syntax_tree(instring, schema)
        return [
            sql_translator.translate(
                root.post_order(), use_bag_semantics, self.builder.grammar.syntax
            )  # type: ignore
            for root in root_list
        ]

    def to_qtree(self, instring, schema={}):
        """
        Translate a relational algebra string into a string representing a
        latex tree, using the grammar.
        """
        root_list = self.to_syntax_tree(instring, schema)
        return qtree_translator.translate(root_list, self.builder.grammar.syntax)  # type: ignore



class BaseSyntax:
    """
    Base syntax class containing common tokens used by all grammars.
    """

    # General tokens.
    terminator = ";"
    delim = ","
    params_start = "_{"
    params_stop = "}"
    paren_left = "("
    paren_right = ")"

    def __init__(self, **kwargs):
        # Set any user defined syntax.
        for key in kwargs:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])


class ConditionSyntax(BaseSyntax):
    """
    Syntax class for condition expressions containing logical and comparison operators.
    """

    # Logical tokens.
    not_op = "not"
    and_op = "and"
    or_op = "or"

    # Comparison operators.
    equal_op = "="
    not_equal_op = "!="
    not_equal_alt_op = "<>"
    less_than_op = "<"
    less_than_equal_op = "<="
    greater_than_op = ">"
    greater_than_equal_op = ">="


class CoreSyntax(ConditionSyntax):
    """
    Syntax class for core relational algebra containing basic operators.
    """

    # Relational algebra operators.
    project_op = "\\project"
    rename_op = "\\rename"
    select_op = "\\select"
    assign_op = ":="
    join_op = "\\join"
    difference_op = "\\difference"
    union_op = "\\union"


class ExtendedSyntax(CoreSyntax):
    """
    Syntax class for extended relational algebra containing additional join operators.
    """

    # Additional join operators.
    theta_join_op = "\\theta_join"
    natural_join_op = "\\natural_join"
    full_outer_join_op = "\\full_outer_join"
    left_outer_join_op = "\\left_outer_join"
    right_outer_join_op = "\\right_outer_join"
    intersect_op = "\\intersect"

    # Defined operator.
    defined_op = "defined"


class DependencySyntax(ExtendedSyntax):
    """
    Syntax class for dependency grammar containing dependency-specific operators.
    """

    # Dependency operators.
    pk_op = "pk"
    mvd_op = "mvd"
    fd_op = "fd"
    inc_equiv_op = "inc="
    inc_subset_op = "inc⊆"


# Backward compatibility alias
Syntax = DependencySyntax



import functools

from rapt2.rapt import Rapt
from rapt2.transformers.sql import sql_translator
from rapt2.treebrd.errors import AttributeReferenceError
from rapt2.treebrd.grammars import CoreGrammar
from rapt2.treebrd.grammars.extended_grammar import ExtendedGrammar
from rapt2.treebrd.treebrd import TreeBRD
from tests.transformers.test_transfomer import TestTransformer


class TestSQL(TestTransformer):
    grammar = CoreGrammar()

    def setUp(self):
        self.translate = self.translate_func(
            functools.partial(
                Rapt(grammar="Extended Grammar").to_sql, use_bag_semantics=True
            )
        )
        self.translate_set = self.translate_func(
            functools.partial(Rapt(grammar="Extended Grammar").to_sql)
        )


class TestRelation(TestSQL):
    def test_single_relation(self):
        ra = "alpha;"
        expected = ["SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_single_relation_set(self):
        ra = "alpha;"
        expected = ["SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3 FROM alpha"]
        actual = self.translate_set(ra)
        self.assertEqual(expected, actual)

    def test_multiple_relations(self):
        ra = "alpha; beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha",
            "SELECT beta.b1, beta.b2, beta.b3 FROM beta",
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestSelect(TestSQL):
    def test_single_select(self):
        ra = r"\select_{a1=a2} alpha;"
        expected = ["SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a1 = a2)"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_single_select_set(self):
        ra = r"\select_{a1=a2} alpha;"
        expected = [
            "SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (a1 = a2)"
        ]
        actual = self.translate_set(ra)
        self.assertEqual(expected, actual)

    def test_single_select_with_relation(self):
        ra = r"\select_{alpha.a1=a2} alpha;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE (alpha.a1 = a2)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_selects(self):
        ra = r"\select_{a1=1} \select_{a2=2} alpha;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha WHERE ((a2 = 2)) AND ((a1 = 1))"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_select_with_precedence(self):
        ra = r"\select_{a2=a3} (\select_{a1=a2} alpha);"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha "
            "WHERE ((a1 = a2)) AND ((a2 = a3))"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_select_with_multiple_conditions(self):
        ra = r"\select_{a1=2 or a1=1} \select_{a2=2 or a2=1} alpha;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha "
            "WHERE (((a2 = 2) OR (a2 = 1))) AND (((a1 = 2) OR (a1 = 1)))"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_single_select_wrong_attrs(self):
        ra = r"\select_{b1=bee_one} Alpha;"
        self.assertRaises(AttributeReferenceError, self.translate, ra)

    def test_multiple_selects_wrong_attrs(self):
        ra = r"\select_{b1=1} \select_{bee_two=2} Alpha;"
        self.assertRaises(AttributeReferenceError, self.translate, ra)

    def test_multiple_selects_wrong_attrs_both(self):
        ra = r"\select_{bee_one=1} \select_{a_two=2} Alpha;"
        self.assertRaises(AttributeReferenceError, self.translate, ra)


class TestProject(TestSQL):
    def test_simple(self):
        ra = r"\project_{a1, a2, a3} alpha;"
        expected = ["SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_prefixed(self):
        ra = r"\project_{alpha.a1, alpha.a2, alpha.a3} alpha;"
        expected = ["SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_subset(self):
        ra = r"\project_{a1, alpha.a2} alpha;"
        expected = ["SELECT alpha.a1, alpha.a2 FROM alpha"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_not_a_subset(self):
        ra = r"\project_{b} Alpha;"
        self.assertRaises(AttributeReferenceError, self.translate, ra)

    def test_not_a_subset2(self):
        ra = r"\project_{Alpha.a1, b} Alpha;"
        self.assertRaises(AttributeReferenceError, self.translate, ra)

    def test_not_a_subset3(self):
        ra = r"\project_{a1, Alpha.b} Alpha;"
        self.assertRaises(AttributeReferenceError, self.translate, ra)


class TestRename(TestSQL):
    def test_relation(self):
        ra = r"\rename_{apex} alpha;"
        expected = [
            "SELECT apex.a1, apex.a2, apex.a3 "
            "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) "
            "AS apex(a1, a2, a3)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_attributes(self):
        ra = r"\rename_{(a, b, c)} alpha;"
        expected = [
            "SELECT alpha.a, alpha.b, alpha.c "
            "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) "
            "AS alpha(a, b, c)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_all(self):
        ra = r"\rename_{apex(a, b, c)} alpha;"
        expected = [
            "SELECT apex.a, apex.b, apex.c "
            "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) "
            "AS apex(a, b, c)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_with_select(self):
        ra = r"\rename_{apex(a, b, c)} \select_{a1 = a2} alpha;"
        expected = [
            "SELECT apex.a, apex.b, apex.c "
            "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha "
            "WHERE (a1 = a2)) AS apex(a, b, c)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_is_projected(self):
        ra = r"\project_{a, c} \rename_{apex(a, b, c)} alpha;"
        expected = [
            "SELECT apex.a, apex.c "
            "FROM (SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) "
            "AS apex(a, b, c)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestAssignment(TestSQL):
    def test_relation(self):
        ra = "newalpha := alpha;"
        expected = [
            "CREATE TEMPORARY TABLE newalpha(a1, a2, a3) AS "
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_relation_with_attributes(self):
        ra = "newalpha(a, b, c) := alpha;"
        expected = [
            "CREATE TEMPORARY TABLE newalpha(a, b, c) AS "
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_relation_with_ambiguous_attributes(self):
        ra = "newalpha(a, b, c) := ambiguous;"
        expected = [
            "CREATE TEMPORARY TABLE newalpha(a, b, c) AS "
            "SELECT ambiguous.a, ambiguous.a, ambiguous.b "
            "FROM ambiguous"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_select(self):
        ra = r"niche := \select_{b1=1} beta;"
        expected = [
            "CREATE TEMPORARY TABLE niche(b1, b2, b3) AS "
            "SELECT beta.b1, beta.b2, beta.b3 FROM beta "
            "WHERE (b1 = 1)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_select_with_attributes(self):
        ra = r"niche(a, b, c) := \select_{b1=1} beta;"
        expected = [
            "CREATE TEMPORARY TABLE niche(a, b, c) AS "
            "SELECT beta.b1, beta.b2, beta.b3 FROM beta "
            "WHERE (b1 = 1)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_project(self):
        ra = "bilk := \\project_{b1, b2} beta;"
        expected = [
            "CREATE TEMPORARY TABLE bilk(b1, b2) AS SELECT beta.b1, beta.b2 FROM beta"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_project_with_attributes(self):
        ra = "bilk(a, b) := \\project_{b1, b2} beta;"
        expected = [
            "CREATE TEMPORARY TABLE bilk(a, b) AS SELECT beta.b1, beta.b2 FROM beta"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_project_select(self):
        ra = "combine := \\project_{b1, b2} \\select_{b1 = 1} beta;"
        expected = [
            "CREATE TEMPORARY TABLE combine(b1, b2) AS "
            "SELECT beta.b1, beta.b2 FROM beta "
            "WHERE (b1 = 1)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_join(self):
        ra = "bound := alpha \\join beta;"
        expected = [
            "CREATE TEMPORARY TABLE bound(a1, a2, a3, b1, b2, b3) AS "
            "SELECT alpha.a1, alpha.a2, alpha.a3, beta.b1, beta.b2, "
            "beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "CROSS JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_join_with_attributes(self):
        ra = "bound(a, b, c, d, e, f) := alpha \\join beta;"
        expected = [
            "CREATE TEMPORARY TABLE bound(a, b, c, d, e, f) AS "
            "SELECT alpha.a1, alpha.a2, alpha.a3, beta.b1, beta.b2, "
            "beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "CROSS JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestJoin(TestSQL):
    def test_relation(self):
        ra = "alpha \\join beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "CROSS JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_three_relations(self):
        self.maxDiff = None
        ra = "alpha \\join beta \\join gamma;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, beta.b1, beta.b2, "
            "beta.b3, gamma.g1, gamma.g2 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "CROSS JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "CROSS JOIN "
            "(SELECT gamma.g1, gamma.g2 FROM gamma) AS gamma"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_relations(self):
        ra = "alpha \\join beta \\join gamma \\join delta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, beta.b1, beta.b2, "
            "beta.b3, gamma.g1, gamma.g2, delta.d1, delta.d2 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "CROSS JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "CROSS JOIN "
            "(SELECT gamma.g1, gamma.g2 FROM gamma) AS gamma "
            "CROSS JOIN "
            "(SELECT delta.d1, delta.d2 FROM delta) AS delta"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_right_join_first(self):
        ra = "alpha \\join (beta \\join gamma);"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, beta.b1, beta.b2, "
            "beta.b3, gamma.g1, gamma.g2 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "CROSS JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "CROSS JOIN "
            "(SELECT gamma.g1, gamma.g2 FROM gamma) AS gamma"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_select_right(self):
        ra = "delta \\join \\select_{g1 = g2} gamma;"
        expected = [
            "SELECT delta.d1, delta.d2, gamma.g1, gamma.g2 FROM "
            "(SELECT delta.d1, delta.d2 FROM delta) AS delta "
            "CROSS JOIN "
            "(SELECT gamma.g1, gamma.g2 FROM gamma "
            "WHERE (g1 = g2)) AS gamma"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_select_left(self):
        ra = "\\select_{g1 = g2} gamma \\join  delta;"
        expected = [
            "SELECT gamma.g1, gamma.g2, delta.d1, delta.d2 FROM "
            "(SELECT gamma.g1, gamma.g2 FROM gamma "
            "WHERE (g1 = g2)) AS gamma "
            "CROSS JOIN "
            "(SELECT delta.d1, delta.d2 FROM delta) AS delta"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_select_on_join(self):
        ra = "\\select_{g1 = g2} (gamma \\join delta);"
        translation = self.translate(ra)
        expected = [
            "SELECT gamma.g1, gamma.g2, delta.d1, delta.d2 FROM "
            "(SELECT gamma.g1, gamma.g2 FROM gamma) AS gamma "
            "CROSS JOIN "
            "(SELECT delta.d1, delta.d2 FROM delta) AS delta "
            "WHERE (g1 = g2)"
        ]
        actual = translation
        self.assertEqual(expected, actual)

    def test_join_with_rename(self):
        ra = "\\rename_{g} gamma \\join \\rename_{h} gamma;"
        translation = self.translate(ra)
        expected = [
            "SELECT g.g1, g.g2, h.g1, h.g2 FROM "
            "(SELECT g.g1, g.g2 FROM (SELECT gamma.g1, gamma.g2 "
            "FROM gamma) AS g(g1, g2)) AS g "
            "CROSS JOIN "
            "(SELECT h.g1, h.g2 FROM (SELECT gamma.g1, gamma.g2 "
            "FROM gamma) AS h(g1, g2)) AS h"
        ]
        actual = translation
        self.assertEqual(expected, actual)

    def test_join_with_rename_set(self):
        ra = "\\rename_{g} gamma \\join \\rename_{h} gamma;"
        translation = self.translate_set(ra)
        expected = [
            "SELECT DISTINCT g.g1, g.g2, h.g1, h.g2 FROM "
            "(SELECT DISTINCT g.g1, g.g2 "
            "FROM (SELECT DISTINCT gamma.g1, gamma.g2 "
            "FROM gamma) AS g(g1, g2)) AS g "
            "CROSS JOIN "
            "(SELECT DISTINCT h.g1, h.g2 "
            "FROM (SELECT DISTINCT gamma.g1, gamma.g2 "
            "FROM gamma) AS h(g1, g2)) AS h"
        ]
        actual = translation
        self.assertEqual(expected, actual)

    def test_join_with_rename_attributes(self):
        ra = "\\rename_{g(a, b)} gamma \\join \\rename_{h(a, b)} gamma;"
        translation = self.translate(ra)
        expected = [
            "SELECT g.a, g.b, h.a, h.b FROM "
            "(SELECT g.a, g.b FROM (SELECT gamma.g1, gamma.g2 "
            "FROM gamma) AS g(a, b)) AS g "
            "CROSS JOIN "
            "(SELECT h.a, h.b FROM (SELECT gamma.g1, gamma.g2 "
            "FROM gamma) AS h(a, b)) AS h"
        ]
        actual = translation
        self.assertEqual(expected, actual)


class TestNaturalJoin(TestSQL):
    grammar = ExtendedGrammar()

    def test_relation_simple(self):
        ra = "alpha \\natural_join alphatwin;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "NATURAL JOIN "
            "(SELECT alphatwin.a1, alphatwin.a2, alphatwin.a3 "
            "FROM alphatwin) AS alphatwin"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_three_relations(self):
        ra = "alpha \\natural_join alphatwin \\natural_join alphaprime;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, alphaprime.a4, "
            "alphaprime.a5 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "NATURAL JOIN "
            "(SELECT alphatwin.a1, alphatwin.a2, alphatwin.a3 "
            "FROM alphatwin) AS alphatwin "
            "NATURAL JOIN "
            "(SELECT alphaprime.a1, alphaprime.a4, alphaprime.a5 "
            "FROM alphaprime) AS alphaprime"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestThetaJoin(TestSQL):
    grammar = ExtendedGrammar()

    def test_relation(self):
        ra = "alpha \\join_{a1 = b1} beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON (a1 = b1)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_theta_join_then_join(self):
        ra = "(alpha \\join_{alpha.a1=beta.b1 and b3>50} beta) \\join gamma;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3, "
            "gamma.g1, gamma.g2 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON ((alpha.a1 = beta.b1) AND (b3 > 50)) "
            "CROSS JOIN (SELECT gamma.g1, gamma.g2 FROM gamma) AS gamma"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestFullOuterJoin(TestSQL):
    grammar = ExtendedGrammar()

    def test_relation(self):
        ra = "alpha \\full_outer_join_{a1 = b1} beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "FULL OUTER JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON (a1 = b1)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_relation_set(self):
        ra = "alpha \\full_outer_join_{a1 = b1} beta;"
        expected = [
            "SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "FULL OUTER JOIN "
            "(SELECT DISTINCT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON (a1 = b1)"
        ]
        actual = self.translate_set(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = "alpha \\full_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "FULL OUTER JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON ((alpha.a1 = beta.b1) AND (b3 > 50))"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestLeftOuterJoin(TestSQL):
    grammar = ExtendedGrammar()

    def test_relation(self):
        ra = "alpha \\left_outer_join_{a1 = b1} beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "LEFT OUTER JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON (a1 = b1)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_relation_set(self):
        ra = "alpha \\left_outer_join_{a1 = b1} beta;"
        expected = [
            "SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "LEFT OUTER JOIN "
            "(SELECT DISTINCT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON (a1 = b1)"
        ]
        actual = self.translate_set(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = "alpha \\left_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "LEFT OUTER JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON ((alpha.a1 = beta.b1) AND (b3 > 50))"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestRightOuterJoin(TestSQL):
    grammar = ExtendedGrammar()

    def test_relation(self):
        ra = "alpha \\right_outer_join_{a1 = b1} beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "RIGHT OUTER JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON (a1 = b1)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_relation_set(self):
        ra = "alpha \\right_outer_join_{a1 = b1} beta;"
        expected = [
            "SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT DISTINCT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "RIGHT OUTER JOIN "
            "(SELECT DISTINCT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON (a1 = b1)"
        ]
        actual = self.translate_set(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = "alpha \\right_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            "SELECT alpha.a1, alpha.a2, alpha.a3, "
            "beta.b1, beta.b2, beta.b3 FROM "
            "(SELECT alpha.a1, alpha.a2, alpha.a3 FROM alpha) AS alpha "
            "RIGHT OUTER JOIN "
            "(SELECT beta.b1, beta.b2, beta.b3 FROM beta) AS beta "
            "ON ((alpha.a1 = beta.b1) AND (b3 > 50))"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestSet:
    __test__ = False

    def test_simple(self):
        ra = "gamma {operator} gammatwin;".format(operator=self.ra_operator)

        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        name = id(root_list[0])
        actual = sql_translator.translate(root_list, use_bag_semantics=True)

        expected = [
            "SELECT g1, g2 FROM ("
            "SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) "
            "AS _{name}".format(operator=self.sql_operator, name=name)
        ]
        self.assertEqual(expected, actual)

    def test_simple_multiple(self):
        ra = "gamma {operator} gammatwin {operator} gammaprime;".format(
            operator=self.ra_operator
        )

        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        root_name = id(root_list[0])
        child_name = id(root_list[0].left)
        actual = sql_translator.translate(root_list, use_bag_semantics=True)

        expected = [
            "SELECT g1, g2 FROM ("
            "SELECT g1, g2 FROM "
            "(SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) AS _{name1} "
            "{operator} ALL "
            "SELECT gammaprime.g1, gammaprime.g2 FROM gammaprime) AS _{name2}".format(
                operator=self.sql_operator, name1=child_name, name2=root_name
            )
        ]
        self.assertEqual(expected, actual)

    def test_select_simple(self):
        ra = "\\select_{{g1 = g2}} (gamma {operator} gammatwin);".format(
            operator=self.ra_operator
        )
        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        name = id(root_list[0].child)
        actual = sql_translator.translate(root_list, use_bag_semantics=True)
        expected = [
            "SELECT g1, g2 FROM "
            "(SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) "
            "AS _{name} WHERE g1 = g2".format(operator=self.sql_operator, name=name)
        ]
        self.assertEqual(expected, actual)

    def test_project_simple(self):
        ra = "\\project_{{g2}} (gamma {operator} gammatwin);".format(
            operator=self.ra_operator
        )
        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        name = id(root_list[0].child)
        actual = sql_translator.translate(root_list, use_bag_semantics=True)
        expected = [
            "SELECT g2 FROM "
            "(SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) "
            "AS _{name}".format(operator=self.sql_operator, name=name)
        ]
        self.assertEqual(expected, actual)

    def test_project_left(self):
        ra = "\\project_{{g1, g2}} gammaprime {operator} gamma;".format(
            operator=self.ra_operator
        )

        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        name = id(root_list[0])
        actual = sql_translator.translate(root_list, use_bag_semantics=True)

        expected = [
            "SELECT g1, g2 FROM "
            "(SELECT gammaprime.g1, gammaprime.g2 FROM gammaprime "
            "{operator} ALL "
            "SELECT gamma.g1, gamma.g2 FROM gamma) AS _{name}".format(
                operator=self.sql_operator, name=name
            )
        ]
        self.assertEqual(expected, actual)

    def test_rename_simple(self):
        ra = "\\rename_{{g}} (gamma {operator} gammatwin);".format(
            operator=self.ra_operator
        )

        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        name = id(root_list[0].child)
        actual = sql_translator.translate(root_list, use_bag_semantics=True)

        expected = [
            "SELECT g.g1, g.g2 FROM "
            "(SELECT g1, g2 FROM (SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) "
            "AS _{name}) AS g(g1, g2)".format(operator=self.sql_operator, name=name)
        ]
        self.assertEqual(expected, actual)

    def test_rename_attributes(self):
        ra = "\\rename_{{g(a, b)}} (gamma {operator} gammatwin);".format(
            operator=self.ra_operator
        )

        root_list = TreeBRD(self.grammar).build(instring=ra, schema=self.schema)
        name = id(root_list[0].child)
        actual = sql_translator.translate(root_list, use_bag_semantics=True)

        expected = [
            "SELECT g.a, g.b FROM "
            "(SELECT g1, g2 FROM (SELECT gamma.g1, gamma.g2 FROM gamma "
            "{operator} ALL "
            "SELECT gammatwin.g1, gammatwin.g2 FROM gammatwin) AS _{name}) "
            "AS g(a, b)".format(operator=self.sql_operator, name=name)
        ]
        self.assertEqual(expected, actual)


class TestUnion(TestSQL, TestSet):
    ra_operator = "\\union"
    sql_operator = "UNION"


class TestDifference(TestSQL, TestSet):
    ra_operator = "\\difference"
    sql_operator = "EXCEPT"


class TestIntersection(TestSQL, TestSet):
    grammar = ExtendedGrammar()
    ra_operator = "\\intersect"
    sql_operator = "INTERSECT"


class TestPrimaryKey(TestSQL):
    def setUp(self):
        from rapt2.rapt import Rapt
        import functools
        self.rapt = Rapt(grammar="Dependency Grammar")
        self.translate = self.translate_func(self.rapt.to_sql)
        self.translate_set = self.translate_func(functools.partial(self.rapt.to_sql, use_bag_semantics=False))

    def test_single_attribute_primary_key(self):
        ra = "pk_{a} R;"
        expected = ["ALTER TABLE r ADD PRIMARY KEY (a)"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_attributes_primary_key(self):
        ra = "pk_{a, b} R;"
        expected = ["ALTER TABLE r ADD PRIMARY KEY (a, b)"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_primary_key_with_different_relation(self):
        ra = "pk_{id, name} Users;"
        expected = ["ALTER TABLE users ADD PRIMARY KEY (id, name)"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_primary_key_set_semantics(self):
        ra = "pk_{a, b} R;"
        expected = ["ALTER TABLE r ADD PRIMARY KEY (a, b)"]
        actual = self.translate_set(ra)
        self.assertEqual(expected, actual)

    def test_multiple_primary_key_statements(self):
        ra = "pk_{a} R; pk_{id} Users;"
        expected = [
            "ALTER TABLE r ADD PRIMARY KEY (a)",
            "ALTER TABLE users ADD PRIMARY KEY (id)"
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)



from pyparsing import ParseResults

from .condition_node import (BinaryConditionalOperator,
                                          BinaryConditionNode, ConditionNode,
                                          IdentityConditionNode,
                                          UnaryConditionalOperator,
                                          UnaryConditionNode)

from .grammars.proto_grammar import ProtoGrammar
from .node import (AssignNode, CrossJoinNode, DefinitionNode, DifferenceNode,
                   FullOuterJoinNode, FunctionalDependencyNode,
                   InclusionEquivalenceNode, InclusionSubsumptionNode,
                   IntersectNode, LeftOuterJoinNode, MultivaluedDependencyNode,
                   NaturalJoinNode, PrimaryKeyNode, ProjectNode, RelationNode,
                   RenameNode, RightOuterJoinNode, SelectNode, ThetaJoinNode,
                   UnionNode)
from .schema import Schema


class TreeBRD:
    """
    A Tree Builder for Relational (Algebra) Data. TreeBRD is a factory class
    that builds forests of relational algebra syntax trees. STARBuilder
    """

    grammar: ProtoGrammar

    def __init__(self, grammar: ProtoGrammar):
        self.grammar = grammar

    def build(self, instring, schema):
        ra = self.grammar.parse(instring)
        _schema = Schema(schema)
        return [self.to_node(statement, _schema) for statement in ra[:]]

    def to_node(self, exp: ParseResults, schema):
        """
        Return a Node that is the root of the parse tree for the the specified
        expression.

        :param exp: A list that represents a relational algebra expression.
        Assumes that this list was generated by pyparsing.
        :param schema: A dictionary of relation names to attribute names used
        for verification and generating attributes.
        :return: A Node.
        """

        # Check if this is a dependency statement
        if self.is_dependency_statement(exp):
            return self.create_dependency_node(exp, schema)

        # A relation node.
        if len(exp) == 1 and isinstance(exp[0], str):
            name = exp[0]
            node = RelationNode(name=name, schema=schema)

        # An expression.
        elif len(exp) == 1 and isinstance(exp[0], ParseResults):
            node = self.to_node(exp[0], schema)

        # Unary operators.
        elif isinstance(exp[0], str) and self.grammar.is_unary(exp[0]):
            child = self.to_node(exp[2:], schema)
            node = self.create_unary_node(
                operator=exp[0], child=child, param=exp[1], schema=schema
            )

        # Definition.
        elif len(exp) == 2 and isinstance(exp[0], str) and isinstance(exp[1], ParseResults):
            name = exp[0]
            attributes = list(exp[1])  # Convert ParseResults to list
            node = DefinitionNode(name=name, attributes=attributes, schema=schema)

        # Assignment.
        elif exp[1] is self.grammar.syntax.assign_op:
            child = self.to_node(exp[2:], schema)
            node = self.create_unary_node(
                operator=exp[1], child=child, param=exp[0], schema=schema
            )

        # Binary operators.
        elif self.grammar.is_binary(exp[1]):
            # Pyparsing will put different operators with the same precedence
            # in the same list. This can be a problem when we mix operators with
            # and without parameters (for example join). We avoid this below
            # and build from right to left, to create the correct syntax tree.
            if isinstance(exp[-2], str):
                # Operator without parameters
                op_pos = -2
                param = None
            else:
                op_pos = -3
                param = exp[-2]

            operator = exp[op_pos]
            left = self.to_node(exp[:op_pos], schema)
            right = self.to_node(exp[-1], schema)
            node = self.create_binary_node(
                operator=operator, left=left, right=right, param=param
            )

        else:
            raise ValueError

        return node

    def create_condition_node(self, conditions: ParseResults) -> ConditionNode:
        if isinstance(conditions, str):
            # Attribute Condition Node
            return IdentityConditionNode(conditions)

        elif len(conditions) == 2:
            # Unary Condition Node
            op, right = conditions
            right_node = self.create_condition_node(right)

            return UnaryConditionNode(
                op=UnaryConditionalOperator.from_syntax(self.grammar.syntax, op),
                child=right_node,
            )

        elif len(conditions) == 3:
            # Binary Condition Node
            left, op, right = conditions
            left_node = self.create_condition_node(left)
            right_node = self.create_condition_node(right)

            return BinaryConditionNode(
                op=BinaryConditionalOperator.from_syntax(self.grammar.syntax, op),
                left=left_node,
                right=right_node,
            )

        raise ValueError

    def create_unary_node(
        self, operator, child, param: ParseResults | None = None, schema=None
    ):
        """
        Return a Unary Node whose type depends on the specified operator.

        :param schema:
        :param child:
        :param operator: A relational algebra operator (see constants.py)
        :param param: A list of parameters for the operator.
        :return: A Unary Node.
        """

        if operator == self.grammar.syntax.select_op:
            condition = self.create_condition_node(param[0])
            node = SelectNode(child, condition)

        elif operator == self.grammar.syntax.project_op:
            node = ProjectNode(child, param)

        elif operator == self.grammar.syntax.rename_op:
            name = None
            attributes = []
            if isinstance(param[0], str):
                name = param.pop(0)
            if param:
                attributes = param[0]
            node = RenameNode(child, name, attributes, schema)

        elif operator == self.grammar.syntax.assign_op:
            name = param[0]
            attributes = [] if len(param) < 2 else param[1]
            node = AssignNode(child, name, attributes, schema)

        else:
            raise ValueError

        return node

    def create_binary_node(
        self, operator, left, right, param: ParseResults | None = None
    ):
        """
        Return a Node whose type depends on the specified operator.

        :param operator: A relational algebra operator (see constants.py)
        :return: A Node.
        """

        # Join operators
        if operator == self.grammar.syntax.join_op:
            node = CrossJoinNode(left, right)

        elif operator == self.grammar.syntax.natural_join_op:
            node = NaturalJoinNode(left, right)

        elif operator == self.grammar.syntax.theta_join_op:
            condition = self.create_condition_node(param[0])
            node = ThetaJoinNode(left, right, condition)

        elif operator == self.grammar.syntax.full_outer_join_op:
            condition = self.create_condition_node(param[0])
            node = FullOuterJoinNode(left, right, condition)

        elif operator == self.grammar.syntax.left_outer_join_op:
            condition = self.create_condition_node(param[0])
            node = LeftOuterJoinNode(left, right, condition)

        elif operator == self.grammar.syntax.right_outer_join_op:
            condition = self.create_condition_node(param[0])
            node = RightOuterJoinNode(left, right, condition)

        # Set operators
        elif operator == self.grammar.syntax.union_op:
            node = UnionNode(left, right)

        elif operator == self.grammar.syntax.difference_op:
            node = DifferenceNode(left, right)

        elif operator == self.grammar.syntax.intersect_op:
            node = IntersectNode(left, right)

        else:
            raise ValueError

        return node

    def is_dependency_statement(self, exp: ParseResults) -> bool:
        """
        Check if the expression is a dependency statement.

        :param exp: A parsed expression
        :return: True if it's a dependency statement
        """
        if not exp or len(exp) < 2:
            return False

        # Check if the syntax has dependency operators
        dependency_ops = []
        if hasattr(self.grammar.syntax, 'pk_op'):
            dependency_ops.append(self.grammar.syntax.pk_op)
        if hasattr(self.grammar.syntax, 'mvd_op'):
            dependency_ops.append(self.grammar.syntax.mvd_op)
        if hasattr(self.grammar.syntax, 'fd_op'):
            dependency_ops.append(self.grammar.syntax.fd_op)
        if hasattr(self.grammar.syntax, 'inc_equiv_op'):
            dependency_ops.append(self.grammar.syntax.inc_equiv_op)
        if hasattr(self.grammar.syntax, 'inc_subset_op'):
            dependency_ops.append(self.grammar.syntax.inc_subset_op)

        return exp[0] in dependency_ops

    def create_dependency_node(self, exp: ParseResults, schema):
        """
        Create a dependency node from a parsed dependency statement.

        :param exp: A parsed dependency expression
        :param schema: A schema object
        :return: A dependency node
        """
        operator = exp[0]

        if operator == self.grammar.syntax.pk_op:
            # pk_{attributes} relation
            attributes = exp[1]
            relation_name = exp[2]
            return PrimaryKeyNode(relation_name, attributes)

        elif operator in (self.grammar.syntax.mvd_op, self.grammar.syntax.fd_op):
            # mvd_{attribute1, attribute2} relation OR mvd_{attribute1, attribute2} \select_{conditions} relation
            # fd_{attribute1, attribute2} relation OR fd_{attribute1, attribute2} \select_{conditions} relation
            return self._create_dependency_with_optional_select(exp, schema, operator)

        elif operator == self.grammar.syntax.inc_equiv_op:
            # inc=_{attributes} (select_or_relation, select_or_relation)
            attributes = exp[1]
            relations_expr = exp[2]
            
            if isinstance(relations_expr[0], str):
                # Simple form: inc=_{attributes} (relation1, relation2)
                relation_names = relations_expr
                left_child = RelationNode(relation_names[0], schema)
                right_child = RelationNode(relation_names[1], schema)
            else:
                # Complex form: inc=_{attributes} (select_or_relation, select_or_relation)
                # relations_expr is [[left_expr], [right_expr]] where each can be relation or [select, conditions, relation]
                left_expr = relations_expr[0]
                right_expr = relations_expr[1]
                left_child = self._create_cond_dep_expr_node(left_expr, schema)
                right_child = self._create_cond_dep_expr_node(right_expr, schema)
                relation_names = [left_child.name, right_child.name]
            return InclusionEquivalenceNode(relation_names, attributes, left_child, right_child)

        elif operator == self.grammar.syntax.inc_subset_op:
            # inc⊆_{attributes} (select_or_relation, select_or_relation)
            attributes = exp[1]
            relations_expr = exp[2]
            
            if isinstance(relations_expr[0], str):
                # Simple form: inc⊆_{attributes} (relation1, relation2)
                relation_names = relations_expr
                left_child = RelationNode(relation_names[0], schema)
                right_child = RelationNode(relation_names[1], schema)
            else:
                # Complex form: inc⊆_{attributes} (select_or_relation, select_or_relation)
                # relations_expr is [[left_expr], [right_expr]] where each can be relation or [select, conditions, relation]
                left_expr = relations_expr[0]
                right_expr = relations_expr[1]
                left_child = self._create_cond_dep_expr_node(left_expr, schema)
                right_child = self._create_cond_dep_expr_node(right_expr, schema)
                relation_names = [left_child.name, right_child.name]
            return InclusionSubsumptionNode(relation_names, attributes, left_child, right_child)

        else:
            raise ValueError(f"Unknown dependency operator: {operator}")

    def _create_dependency_with_optional_select(self, exp, schema, operator):
        """
        Create a dependency node (MVD or FD) with optional select conditions.
        
        :param exp: Parsed expression containing attributes and optional select
        :param schema: Schema for relation validation
        :param operator: The dependency operator (mvd_op or fd_op)
        :return: MultivaluedDependencyNode or FunctionalDependencyNode
        """
        attributes = exp[1]
        if len(exp) == 3:
            # Simple form: {operator}_{attributes} relation
            relation_name = exp[2]
            relation_node = RelationNode(relation_name, schema)
            if operator == self.grammar.syntax.mvd_op:
                return MultivaluedDependencyNode(relation_name, attributes, relation_node)
            else:  # fd_op
                return FunctionalDependencyNode(relation_name, attributes, relation_node)
        else:
            # With conditions: {operator}_{attributes} \select_{conditions} relation
            # exp[2] = "\select", exp[3] = conditions, exp[4] = relation_name
            conditions = exp[3]
            relation_name = exp[4]
            condition_node = self.create_condition_node(conditions[0])
            base_relation = RelationNode(relation_name, schema)
            select_node = SelectNode(base_relation, condition_node)
            if operator == self.grammar.syntax.mvd_op:
                return MultivaluedDependencyNode(relation_name, attributes, select_node)
            else:  # fd_op
                return FunctionalDependencyNode(relation_name, attributes, select_node)

    def _create_cond_dep_expr_node(self, expr, schema):
        """
        Create a node from a cond_dep_expr (either relation_name or select with relation).
        
        :param expr: Parsed expression (either relation_name or [select, conditions, relation_name])
        :param schema: Schema for relation validation
        :return: RelationNode or SelectNode
        """
        if isinstance(expr, str):
            # Simple relation name
            return RelationNode(expr, schema)
        elif len(expr) == 1 and isinstance(expr[0], str):
            # Simple relation name wrapped in a list
            return RelationNode(expr[0], schema)
        else:
            # Select with conditions: [select, conditions, relation_name]
            conditions = expr[1]
            relation_name = expr[2]
            condition_node = self.create_condition_node(conditions[0])
            base_relation = RelationNode(relation_name, schema)
            return SelectNode(base_relation, condition_node)



from typing import Union
from ...treebrd.node import (Operator, AssignNode, CrossJoinNode, 
                            DifferenceNode, FullOuterJoinNode,
                            IntersectNode, LeftOuterJoinNode,
                            NaturalJoinNode, PrimaryKeyNode, ProjectNode, RelationNode,
                            RenameNode, RightOuterJoinNode, SelectNode, ThetaJoinNode,
                            UnionNode)
from ...treebrd.condition_node import (BinaryConditionNode, ConditionNode,
                                       IdentityConditionNode, UnaryConditionNode,
                                       BinaryConditionalOperator, UnaryConditionalOperator)
from ..base_translator import BaseTranslator


class SQLQuery:
    """
    Structure defining the building blocks of a SQL query.
    """

    def __init__(self, select_block, from_block, where_block=""):
        self.prefix = ""
        self.select_block = select_block
        self.from_block = from_block
        self.where_block = where_block

    @property
    def _basic_query(self):
        if self.select_block:
            return "{prefix}SELECT {select} FROM {relation}"
        else:
            return "{prefix}{relation}"

    @property
    def _sql_query_skeleton(self):
        sql = self._basic_query
        if self.where_block:
            sql += " WHERE {conditions}"
        return sql

    def to_sql(self):
        """
        Construct a SQL query based on the stored blocks.
        :return: a SQL query
        """
        return self._sql_query_skeleton.format(
            prefix=self.prefix,
            select=self.select_block,
            relation=self.from_block,
            conditions=self.where_block,
        )


class SQLSetQuery(SQLQuery):
    """
    Structure defining the building blocks of a SQL query with set semantics.
    """

    @property
    def _basic_query(self):
        return "{prefix}SELECT DISTINCT {select} FROM {relation}"


class SQLAlterTableQuery(SQLQuery):
    """
    Structure defining an ALTER TABLE SQL statement.
    """

    def __init__(self, table_name, action, attributes):
        """
        Initialize an ALTER TABLE query with table name, action, and attributes.
        :param table_name: The name of the table to alter
        :param action: The alteration action (e.g., "ADD PRIMARY KEY")
        :param attributes: The attributes involved in the action (as a string)
        """
        super().__init__(select_block="", from_block="")
        self.table_name = table_name
        self.action = action
        self.attributes = attributes

    def to_sql(self):
        """
        Return the ALTER TABLE statement directly.
        :return: The ALTER TABLE statement
        """
        return f"ALTER TABLE {self.table_name} {self.action} ({self.attributes})"


class SQLTranslator(BaseTranslator):
    """
    A Translator defining the operations for translating a relational algebra
    statement into a SQL statement using bag semantics.
    """

    query = SQLQuery

    @classmethod
    def _get_temp_name(cls, node):
        return node.name or "_{}".format(id(node))

    def _get_sql_operator(self, node: Union[NaturalJoinNode, ThetaJoinNode, CrossJoinNode, 
                                           FullOuterJoinNode, LeftOuterJoinNode, RightOuterJoinNode,
                                           UnionNode, DifferenceNode, IntersectNode]) -> str:
        operators = {
            Operator.union: "UNION",
            Operator.difference: "EXCEPT",
            Operator.intersect: "INTERSECT",
            Operator.cross_join: "CROSS JOIN",
            Operator.theta_join: "JOIN",
            Operator.natural_join: "NATURAL JOIN",
            Operator.full_outer_join: "FULL OUTER JOIN",
            Operator.left_outer_join: "LEFT OUTER JOIN",
            Operator.right_outer_join: "RIGHT OUTER JOIN",
        }
        return operators[node.operator]

    def relation(self, node: RelationNode) -> SQLQuery:
        """
        Translate a relation node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self.query(select_block=str(node.attributes), from_block=node.name)

    def select(self, node: SelectNode) -> SQLQuery:
        """
        Translate a select node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """

        child_object = self.translate(node.child)
        where_block = self.translate_condition(node.conditions)
        if child_object.where_block:
            where_block = "({0}) AND ({1})".format(
                child_object.where_block, self.translate_condition(node.conditions)
            )
        child_object.where_block = where_block
        if not child_object.select_block:
            child_object.select_block = str(node.attributes)
        return child_object

    def project(self, node: ProjectNode) -> SQLQuery:
        """
        Translate a project node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        child_object = self.translate(node.child)
        child_object.select_block = str(node.attributes)
        return child_object

    def rename(self, node: RenameNode) -> SQLQuery:
        """
        Translate a rename node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        child_object = self.translate(node.child)
        from_block = "({child}) AS {name}({attributes})".format(
            child=child_object.to_sql(),
            name=node.name,
            attributes=", ".join(node.attributes.names),
        )
        return self.query(str(node.attributes), from_block=from_block)

    def assign(self, node: AssignNode) -> SQLQuery:
        """
        Translate an assign node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        child_object = self.translate(node.child)
        child_object.prefix = "CREATE TEMPORARY TABLE {name}({attributes}) AS ".format(
            name=node.name, attributes=", ".join(node.attributes.names)
        )
        return child_object

    def definition(self, node) -> None:
        """
        Ignore definition nodes during SQL translation.
        :param node: a DefinitionNode
        :return: None
        """
        return None

    def natural_join(self, node: NaturalJoinNode) -> SQLQuery:
        """
        Translate an assign node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def theta_join(self, node: ThetaJoinNode) -> SQLQuery:
        """
        Translate an assign node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def cross_join(self, node: CrossJoinNode) -> SQLQuery:
        """
        Translate a cross join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def full_outer_join(self, node: FullOuterJoinNode) -> SQLQuery:
        """
        Translate a full outer join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def left_outer_join(self, node: LeftOuterJoinNode) -> SQLQuery:
        """
        Translate a left outer join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def right_outer_join(self, node: RightOuterJoinNode) -> SQLQuery:
        """
        Translate a right outer join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def union(self, node: UnionNode) -> SQLQuery:
        """
        Translate a union node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._set_op(node)

    def intersect(self, node: IntersectNode) -> SQLQuery:
        """
        Translate an intersection node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._set_op(node)

    def difference(self, node: DifferenceNode) -> SQLQuery:
        """
        Translate an difference node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._set_op(node)

    def _join_helper(self, node: Union[RelationNode, SelectNode, ProjectNode, RenameNode, 
                                      NaturalJoinNode, ThetaJoinNode, CrossJoinNode,
                                      FullOuterJoinNode, LeftOuterJoinNode, RightOuterJoinNode,
                                      UnionNode, DifferenceNode, IntersectNode]) -> str:
        sobject = self.translate(node)
        if node.operator in {
            Operator.cross_join,
            Operator.natural_join,
            Operator.theta_join,
            Operator.full_outer_join,
            Operator.left_outer_join,
            Operator.right_outer_join,
        }:
            return sobject.from_block
        else:
            return "({subquery}) AS {name}".format(
                subquery=sobject.to_sql(), name=self._get_temp_name(node)
            )

    def _join(self, node: Union[NaturalJoinNode, ThetaJoinNode, CrossJoinNode,
                               FullOuterJoinNode, LeftOuterJoinNode, RightOuterJoinNode]) -> SQLQuery:
        """
        Translate a join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """

        select_block = str(node.attributes)
        from_block = "{left} {operator} {right}".format(
            left=self._join_helper(node.left),
            right=self._join_helper(node.right),
            operator=self._get_sql_operator(node),
        )

        if node.operator in {
            Operator.theta_join,
            Operator.full_outer_join,
            Operator.left_outer_join,
            Operator.right_outer_join,
        }:
            # Only theta joins and outer joins have conditions
            if hasattr(node, 'conditions') and getattr(node, 'conditions', None) is not None:
                conditions = getattr(node, 'conditions')
                from_block = "{from_block} ON {conditions}".format(
                    from_block=from_block, conditions=self.translate_condition(conditions)
                )

        return self.query(select_block, from_block, "")

    def _set_op(self, node: Union[UnionNode, DifferenceNode, IntersectNode]) -> SQLQuery:
        """
        Translate a set operator node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        select_block = str(node.attributes)
        from_block = "({left} {operator} ALL {right}) AS {name}".format(
            left=self.translate(node.left).to_sql(),
            right=self.translate(node.right).to_sql(),
            operator=self._get_sql_operator(node),
            name=self._get_temp_name(node),
        )
        return self.query(select_block=select_block, from_block=from_block)

    def identity_condition(self, node: IdentityConditionNode) -> str:
        """
        Translate an identity condition node into SQL.
        :param node: an identity condition node
        :return: SQL representation of the condition
        """
        return node.ident

    def unary_condition(self, node: UnaryConditionNode) -> str:
        """
        Translate a unary condition node into SQL.
        :param node: a unary condition node
        :return: SQL representation of the condition
        """
        match node.op:
            case UnaryConditionalOperator.NOT:
                return f"NOT {self.translate_condition(node.child)}"
            case UnaryConditionalOperator.DEFINED:
                return f"{self.translate_condition(node.child)} IS NOT NULL"
            case _:
                raise ValueError

    def binary_condition(self, node: BinaryConditionNode) -> str:
        """
        Translate a binary condition node into SQL.
        :param node: a binary condition node
        :return: SQL representation of the condition
        """
        match node.op:
            case BinaryConditionalOperator.AND:
                return f"({self.translate_condition(node.left)} AND {self.translate_condition(node.right)})"
            case BinaryConditionalOperator.OR:
                return f"({self.translate_condition(node.left)} OR {self.translate_condition(node.right)})"
            case BinaryConditionalOperator.EQUAL:
                return f"({self.translate_condition(node.left)} = {self.translate_condition(node.right)})"
            case BinaryConditionalOperator.NOT_EQUAL:
                return f"({self.translate_condition(node.left)} != {self.translate_condition(node.right)})"
            case BinaryConditionalOperator.LESS_THAN:
                return f"({self.translate_condition(node.left)} < {self.translate_condition(node.right)})"
            case BinaryConditionalOperator.LESS_THAN_EQUAL:
                return f"({self.translate_condition(node.left)} <= {self.translate_condition(node.right)})"
            case BinaryConditionalOperator.GREATER_THAN:
                return f"({self.translate_condition(node.left)} > {self.translate_condition(node.right)})"
            case BinaryConditionalOperator.GREATER_THAN_EQUAL:
                return f"({self.translate_condition(node.left)} >= {self.translate_condition(node.right)})"
            case _:
                raise ValueError

    def primary_key(self, node: PrimaryKeyNode) -> SQLAlterTableQuery:
        """
        Translate a primary key dependency node into SQL.
        :param node: a PrimaryKeyNode
        :return: a SQLAlterTableQuery object representing the primary key constraint
        """
        # Convert attributes to comma-separated string
        if hasattr(node.attributes, '__iter__') and not isinstance(node.attributes, str):
            attributes_str = ", ".join(str(attr) for attr in node.attributes)
        else:
            attributes_str = str(node.attributes)
        
        return SQLAlterTableQuery(
            table_name=node.relation_name,
            action="ADD PRIMARY KEY",
            attributes=attributes_str
        )

    def multivalued_dependency(self, node) -> None:
        """
        Ignore multivalued dependency nodes during SQL translation.
        :param node: a MultivaluedDependencyNode
        :return: None
        """
        # TODO: Implement multivalued dependency translation
        return None

    def functional_dependency(self, node) -> None:
        """
        Ignore functional dependency nodes during SQL translation.
        :param node: a FunctionalDependencyNode
        :return: None
        """
        # TODO: Implement functional dependency translation
        return None

    def inclusion_equivalence(self, node) -> None:
        """
        Ignore inclusion equivalence nodes during SQL translation.
        :param node: an InclusionEquivalenceNode
        :return: None
        """
        # TODO: Implement inclusion equivalence translation
        return None

    def inclusion_subsumption(self, node) -> None:
        """
        Ignore inclusion subsumption nodes during SQL translation.
        :param node: an InclusionSubsumptionNode
        :return: None
        """
        # TODO: Implement inclusion subsumption translation
        return None

    def translate_condition(self, condition: Union[ConditionNode, BinaryConditionNode, 
                                                 UnaryConditionNode, IdentityConditionNode]) -> str:
        """
        Translate a condition node into SQL.
        :param condition: a condition node
        :return: SQL representation of the condition
        """
        if isinstance(condition, IdentityConditionNode):
            return self.identity_condition(condition)
        elif isinstance(condition, UnaryConditionNode):
            return self.unary_condition(condition)
        elif isinstance(condition, BinaryConditionNode):
            return self.binary_condition(condition)
        else:
            raise ValueError(f"Unknown condition node type: {type(condition)}")


class SetTranslator(SQLTranslator):
    """
    A Translator defining the operations for translating a relational algebra
    statement into a SQL statement using set semantics.
    """

    query = SQLSetQuery

    def _set_op(self, node: Union[UnionNode, DifferenceNode, IntersectNode]) -> SQLSetQuery:
        """
        Translate a set operator node into SQLQuery, using set semantics.
        :param node: a treebrd node
        :return: a SQLSetQuery object for the tree rooted at node
        """
        select_block = str(node.attributes)
        from_block = "({left} {operator} {right}) AS {name}".format(
            left=self.translate(node.left).to_sql(),
            right=self.translate(node.right).to_sql(),
            operator=self._get_sql_operator(node),
            name=self._get_temp_name(node),
        )
        return self.query(select_block=select_block, from_block=from_block)


def translate(root_list, use_bag_semantics=False, syntax=None):
    """
    Translate a list of relational algebra trees into SQL statements.

    :param root_list: a list of tree roots
    :param use_bag_semantics: flag for using relational algebra bag semantics
    :param syntax: syntax instance for custom operators
    :return: a list of SQL statements
    """
    translator = SQLTranslator(syntax) if use_bag_semantics else SetTranslator(syntax)  # type: ignore
    results = []
    for root in root_list:
        translated = translator.translate(root)
        if translated is not None:
            results.append(translated.to_sql())
    return results
